from pytest import mark, raises
from yaml import SafeLoader, load

from otoole.exceptions import OtooleConfigFileError
from otoole.utils import UniqueKeyLoader, validate_config


class TestValidConfigs:
    valid_set = """
        SET_NAME:
          dtype: str
          type: set
          short_name: SET
    """

    valid_parameter = """
        Parameter_Name:
          indices: [SET]
          type: param
          dtype: float
          default: 0
          short_name: Parameter
        SET:
          dtype: str
          type: set
    """

    valid_result = """
        Result_Name:
          indices: [SET]
          type: result
          dtype: float
          default: 0
          calculated: True
          short_name: Result
        SET:
          dtype: str
          type: set
    """

    config_data = [valid_set, valid_parameter, valid_result]

    config_data_ids = ["valid_set", "valid_parameter", "valid_result"]

    @mark.parametrize("config_data", config_data, ids=config_data_ids)
    def test_valid_configs(self, config_data):
        config = load(config_data, Loader=SafeLoader)
        validate_config(config)


class TestInvalidConfigs:
    invalid_duplicate_set_names = """
        SET:
          dtype: float
          type: set
        SET:
          dtype: str
          type: set
    """

    invalid_duplicate_param_names = """
        Parameter:
          indices: [SET]
          type: param
          dtype: float
          default: 0
        Parameter:
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_duplicate_result_names = """
        Result:
          indices: [SET]
          type: result
          dtype: float
          default: 0
          calculated: True
        Result:
          indices: [SET]
          type: result
          dtype: float
          default: 0
          calculated: True
        SET:
          dtype: str
          type: set
    """

    invalid_duplicate_names_diff_types_1 = """
        Parameter:
          indices: [SET]
          type: param
          dtype: float
          default: 0
        Parameter:
          dtype: str
          type: set
    """

    invalid_duplicate_names_diff_types_2 = """
        Parameter:
          indices: [SET]
          type: param
          dtype: float
          default: 0
        PARAMETER:
          dtype: str
          type: set
    """

    config_data_duplicates = [
        invalid_duplicate_set_names,
        invalid_duplicate_param_names,
        invalid_duplicate_result_names,
        invalid_duplicate_names_diff_types_1,
        invalid_duplicate_names_diff_types_2,
    ]

    config_data_duplicates_ids = [
        "invalid_duplicate_set_names",
        "invalid_duplicate_param_names",
        "invalid_duplicate_result_names",
        "invalid_duplicate_names_diff_types_1",
        "invalid_duplicate_names_diff_types_2",
    ]

    @mark.parametrize(
        "config_data", config_data_duplicates, ids=config_data_duplicates_ids
    )
    def test_invalid_config_multi_defs(self, config_data):
        with raises(ValueError):
            load(config_data, Loader=UniqueKeyLoader)


class TestUserDefinedValue:
    invalid_set_type = """
        SET:
          dtype: str
          type: not_valid_type
    """

    invalid_parameter_type = """
        Parameter:
          indices: [SET]
          type: not_valid_type
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_result_type = """
        Result:
          indices: [SET]
          type: not_valid_type
          dtype: float
          default: 0
          calculated: True
        SET:
          dtype: str
          type: set
    """

    invalid_name_spaces = """
        Parameter With Spaces:
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_name_numbers = """
        Parameter12345:
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    # does not inlcude underscores
    invalid_name_special_chars = """
        Parameter!@?%[):
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_name_length = """
        ParameterNameLongerThanThirtyOneChars:
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_short_name_spaces = """
        Parameter:
          short_name: Parameter Short Name
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_short_name_numbers = """
        Parameter:
          short_name: ParameterShortName12345
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    # does not inlcude underscores
    invalid_short_name_special_chars = """
        Parameter:
          short_name: ParameterShortName!@?%[)
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_short_name_length_1 = """
        Parameter:
          short_name: ParameterShortNameLongerThanThirtyOneChars
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_short_name_length_2 = """
        ParameterNameLongerThanThirtyOneChars:
          short_name: ParameterShortNameLongerThanThirtyOneChars
          indices: [SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    config_data_invalid = [
        invalid_set_type,
        invalid_parameter_type,
        invalid_result_type,
        invalid_name_spaces,
        invalid_name_numbers,
        invalid_name_special_chars,
        invalid_name_length,
        invalid_short_name_spaces,
        invalid_short_name_numbers,
        invalid_short_name_special_chars,
        invalid_short_name_length_1,
        invalid_short_name_length_2,
    ]

    config_data_invalid_ids = [
        "invalid_set_type",
        "invalid_parameter_type",
        "invalid_result_type",
        "invalid_name_spaces",
        "invalid_name_numbers",
        "invalid_name_special_chars",
        "invalid_name_length",
        "invalid_short_name_spaces",
        "invalid_short_name_numbers",
        "invalid_short_name_special_chars",
        "invalid_short_name_length_1",
        "invalid_short_name_length_2",
    ]

    @mark.parametrize("config_data", config_data_invalid, ids=config_data_invalid_ids)
    def test_invalid_configs(self, config_data):
        config = load(config_data, Loader=UniqueKeyLoader)
        with raises(OtooleConfigFileError):
            validate_config(config)


class TestUserDefinedSet:
    invalid_dtype = """
        SET:
          dtype: float
          type: set
    """

    invalid_unexpected_field = """
        SET:
          dtype: str
          type: set
          fieldnotexpected: True
    """

    invalid_missing_field = """
        SET:
          type: set
    """

    config_data = [
        invalid_dtype,
        invalid_unexpected_field,
        invalid_missing_field,
    ]

    config_data_ids = [
        "invalid_dtype",
        "invalid_unexpected_field",
        "invalid_missing_field",
    ]

    @mark.parametrize("config_data", config_data, ids=config_data_ids)
    def test_invalid_config_sets(self, config_data):
        config = load(config_data, Loader=UniqueKeyLoader)
        with raises(OtooleConfigFileError):
            validate_config(config)


class TestUserDefinedParameter:
    invalid_dtype = """
        Parameter:
          indices: [SET]
          type: param
          dtype: dict
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_unexpected_field = """
        Parameter:
          indices: [SET]
          type: param
          dtype: float
          default: 0
          fieldnotexpected: True
        SET:
          dtype: str
          type: set
    """

    invalid_missing_field = """
        Parameter:
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_missing_index = """
        Parameter:
          indices: [SET, MISSING_SET]
          type: param
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_default = """
        Parameter:
          indices: [SET, MISSING_SET]
          type: param
          dtype: float
          default: s
        SET:
          dtype: str
          type: set
    """

    config_data = [
        invalid_dtype,
        invalid_unexpected_field,
        invalid_missing_field,
        invalid_missing_index,
        invalid_default,
    ]
    config_data_ids = [
        "invalid_dtype",
        "invalid_unexpected_field",
        "invalid_missing_field",
        "invalid_missing_index",
        "invalid_default",
    ]

    @mark.parametrize("config_data", config_data, ids=config_data_ids)
    def test_invalid_config_params(self, config_data):
        config = load(config_data, Loader=UniqueKeyLoader)
        with raises(OtooleConfigFileError):
            validate_config(config)


class TestInvalidConfigResults:
    invalid_dtype = """
        Result:
          type: param
          dtype: dict
          default: 0
          fieldnotexpected: True
        SET:
          dtype: str
          type: set
    """

    invalid_unexpected_field = """
        Result:
          indices: [SET]
          type: result
          dtype: float
          default: 0
          calculated: True
          fieldnotexpected: True
        SET:
          dtype: str
          type: set
    """

    invalid_missing_field = """
        Result:
          indices: [SET]
          type: result
          dtype: float
          default: 0
        SET:
          dtype: str
          type: set
    """

    invalid_missing_index = """
        Result:
          indices: [SET, MISSING_SET]
          type: result
          dtype: float
          default: 0
          calculated: True
        SET:
          dtype: str
          type: set
    """

    invalid_default = """
        Result:
          indices: [SET]
          type: result
          dtype: float
          default: s
          calculated: True
        SET:
          dtype: str
          type: set
    """

    config_data = [
        invalid_dtype,
        invalid_unexpected_field,
        invalid_missing_field,
        invalid_missing_index,
        invalid_default,
    ]
    config_data_ids = [
        "invalid_dtype",
        "invalid_unexpected_field",
        "invalid_missing_field",
        "invalid_missing_index",
        "invalid_default",
    ]

    @mark.parametrize("config_data", config_data, ids=config_data_ids)
    def test_invalid_config_results(self, config_data):
        config = load(config_data, Loader=UniqueKeyLoader)
        with raises(OtooleConfigFileError):
            validate_config(config)



import os
from subprocess import run
from tempfile import NamedTemporaryFile, mkdtemp

from pytest import mark

from otoole import __version__


class TestConvert:
    def test_version(self):
        result = run(["otoole", "--version"], capture_output=True)
        assert result.stdout.strip().decode() == str(__version__)

    temp = mkdtemp()
    temp_excel = NamedTemporaryFile(suffix=".xlsx")
    temp_datafile = NamedTemporaryFile(suffix=".dat")
    simplicity = os.path.join("tests", "fixtures", "simplicity.txt")
    config_path = os.path.join("tests", "fixtures", "config.yaml")

    test_data = [
        (["otoole", "-v", "convert", "--help"], "usage: otoole convert [-h]"),
        (
            [
                "otoole",
                "-v",
                "convert",
                "datafile",
                "excel",
                simplicity,
                temp_excel.name,
                config_path,
            ],
            "",
        ),
        (
            [
                "otoole",
                "-v",
                "convert",
                "datafile",
                "datafile",
                simplicity,
                temp_datafile.name,
                config_path,
            ],
            "",
        ),
    ]

    @mark.parametrize("commands,expected", test_data, ids=["help", "excel", "datafile"])
    def test_convert_commands(self, commands, expected):
        actual = run(commands, capture_output=True)
        assert expected in str(actual.stdout)
        print(" ".join(commands))
        assert actual.returncode == 0, print(actual.stdout)

    test_errors = [
        (
            ["otoole", "convert", "datafile", "cplex", "x", "y", config_path],
            "error: argument to_format: invalid choice: 'cplex'",
        )
    ]

    @mark.parametrize("commands,expected", test_errors, ids=["invalid"])
    def test_convert_error(self, commands, expected):
        actual = run(commands, capture_output=True)
        assert expected in str(actual.stderr)

    def test_convert_datafile_datafile_no_user_config(self):
        simplicity = os.path.join("tests", "fixtures", "simplicity.txt")
        temp_datafile = NamedTemporaryFile(suffix=".dat")
        commands = [
            "otoole",
            "convert",
            "datafile",
            "datafile",
            simplicity,
            temp_datafile.name,
        ]
        actual = run(commands, capture_output=True)
        assert actual.returncode == 2

    def test_convert_datafile_datafile_with_user_config(self):
        simplicity = os.path.join("tests", "fixtures", "simplicity.txt")
        user_config = os.path.join("tests", "fixtures", "config.yaml")
        temp_datafile = NamedTemporaryFile(suffix=".dat")
        commands = [
            "otoole",
            "-vvv",
            "convert",
            "datafile",
            "datafile",
            simplicity,
            temp_datafile.name,
            user_config,
        ]
        actual = run(commands, capture_output=True)
        assert actual.returncode == 0

    def test_convert_datafile_datafile_with_default_flag(self):
        simplicity = os.path.join("tests", "fixtures", "simplicity.txt")
        user_config = os.path.join("tests", "fixtures", "config.yaml")
        temp_datafile = NamedTemporaryFile(suffix=".dat")
        commands = [
            "otoole",
            "-vvv",
            "convert",
            "datafile",
            "datafile",
            simplicity,
            temp_datafile.name,
            user_config,
            "--write_defaults",
        ]
        actual = run(commands, capture_output=True)
        assert actual.returncode == 0


class TestSetup:

    temp = mkdtemp()
    temp_config = NamedTemporaryFile(suffix=".yaml")

    test_data = [
        (
            [
                "otoole",
                "-v",
                "setup",
                "config",
                NamedTemporaryFile(suffix=".yaml").name,
            ],
            "",
        ),
        (["otoole", "-v", "setup", "config", temp_config.name, "--overwrite"], ""),
    ]

    @mark.parametrize(
        "commands,expected", test_data, ids=["setup", "setup_with_overwrite"]
    )
    def test_setup_commands(self, commands, expected):
        actual = run(commands, capture_output=True)
        assert expected in str(actual.stdout)
        print(" ".join(commands))
        assert actual.returncode == 0, print(actual.stdout)

    test_errors = [
        (["otoole", "-v", "setup", "config", temp_config.name], "OtooleSetupError"),
    ]

    @mark.parametrize("commands,expected", test_errors, ids=["setup_fails"])
    def test_setup_error(self, commands, expected):
        actual = run(commands, capture_output=True)
        assert expected in str(actual.stderr)



"""Ensures that technology and fuel names match the convention

For example, to validate the following list of names, you would use the
config shown below::

    theseUNIQUE_ENTRY1
    are__UNIQUE_ENTRY2
    all__UNIQUE_ENTRY1
    validUNIQUE_ENTRY2
    entryUNIQUE_ENTRY1
    in__UNIQUE_ENTRY2
    a____UNIQUE_ENTRY1
    list_UNIQUE_ENTRY2

Create a yaml validation config with the following format::

    codes:
      some_valid_codes:
        UNIQUE_ENTRY1: Description of unique entry 1
        UNIQUE_ENTRY2: Description of unique entry 2
    schema:
      schema_name:
      - name: first_entry_in_schema
        valid: ['these', 'are__', 'all__', 'valid', 'entry', 'in__', 'a____', 'list_']
        position: (1, 5) # a tuple representing the start and end position
      - name: second_entry_in_schema
        valid: some_valid_codes  # references an entry in the codes section of the
        config
        position: (6, 19) # a tuple representing the start and end position

"""

import logging
import re
from collections import defaultdict
from typing import Dict, List, Sequence

import networkx.algorithms.isolate as isolate
import pandas as pd

from otoole.utils import get_packaged_resource, read_packaged_file
from otoole.visualise.res import create_graph

logger = logging.getLogger(__name__)


def read_validation_config():
    return read_packaged_file("validate.yaml", "otoole")


def check_for_duplicates(codes: Sequence) -> bool:
    duplicate_values = len(codes) != len(set(codes))
    return duplicate_values


def create_schema(config: Dict[str, Dict] = None) -> Dict:
    """Populate the dict of schema with codes from the validation config

    Arguments
    ---------
    config : dict, default=None
        A configuration dictionary containing ``codes`` and ``schema`` keys
    """
    if config is None:
        config = read_validation_config()

    for resource_name, resource_schemas in config["schema"].items():
        logger.debug("%s", resource_name)
        for schema in resource_schemas:

            for items in schema["items"]:  # typing: List

                if isinstance(items["valid"], str):
                    items["valid"] = list(
                        config["codes"][items["valid"]].keys()
                    )  # typing: List
                    logger.debug("create_schema: %s", items["valid"])
                elif isinstance(items["valid"], list):
                    pass
                else:
                    raise ValueError("Entry {} is not correct".format(schema["name"]))

                if check_for_duplicates(items["valid"]):
                    raise ValueError(
                        "There are duplicate values in codes for {}", schema["name"]
                    )

    return config["schema"]


def compose_expression(schema: List) -> str:
    """Generates a regular expression from a schema

    Returns
    -------
    str
    """
    expression = "^"
    for x in schema:
        logger.debug("compose_expression: %s", x["valid"])
        valid_entries = "|".join(x["valid"])
        expression += "({})".format(valid_entries)
    return expression


def compose_multi_expression(resource: List) -> str:
    """Concatenates multiple expressions using an OR operator

    Use to validate elements using an OR operation e.g. the elements
    must match this expression OR the expression
    """
    expressions = []
    for schemas in resource:
        expressions.append(compose_expression(schemas["items"]))
    return "|".join(expressions)


def validate(expression: str, name: str) -> bool:
    """Determine if ``name`` matches the ``expression``

    Arguments
    ---------
    expression : str
    name : str

    Returns
    -------
    bool
    """
    logger.debug("Running validation for %s", name)

    valid = False

    pattern = re.compile(expression)

    if pattern.fullmatch(name):
        valid = True
    else:
        valid = False
    return valid


def validate_resource(
    input_data: Dict[str, pd.DataFrame], resource: str, schemas: List[Dict]
) -> None:
    """Validates a single resource against the validation config.

    Arguments
    ---------
    input_data: dict[str,pd.DataFrame]
        otoole internal datastore
    resource: str
    schemas : List[Dict]
        The schema from which to create a validation expression
    """

    print(
        "Validating {} with {}\n".format(
            resource, ", ".join([x["name"] for x in schemas])
        )
    )

    logger.debug(schemas)

    expression = compose_multi_expression(schemas)
    resources = get_packaged_resource(input_data, resource)

    valid_names = []
    invalid_names = []

    for row in resources:
        name = row["VALUE"]
        valid = validate(expression, row["VALUE"])
        if valid:
            valid_names.append(name)
        else:
            invalid_names.append(name)

    if invalid_names:
        msg = "{} invalid names:\n{}\n"
        print(msg.format(len(invalid_names), ", ".join(invalid_names)))
    if valid_names:
        msg = "{} valid names:\n{}\n"
        print(msg.format(len(valid_names), ", ".join(valid_names)))


def identify_orphaned_fuels_techs(
    input_data: Dict[str, pd.DataFrame]
) -> Dict[str, str]:
    """Returns a list of fuels and technologies which are unconnected

    Returns
    -------
    dict

    """
    graph = create_graph(input_data)

    number_of_isolates = isolate.number_of_isolates(graph)
    logger.debug("There are {} isolated nodes in the graph".format(number_of_isolates))

    isolated_nodes: Dict = defaultdict(list)

    for node_name in list(isolate.isolates(graph)):
        node_data = graph.nodes[node_name]
        isolated_nodes[node_data["type"]].append(node_name)

    return isolated_nodes


def main(input_data: Dict[str, pd.DataFrame], config=None):

    print("\n***Beginning validation***\n")
    schema = create_schema(config)

    for resource, schemas in schema.items():
        validate_resource(input_data, resource, schemas)

    print("\n***Checking graph structure***")
    isolated_nodes = identify_orphaned_fuels_techs(input_data)

    msg = ""
    for node_type, node_names in isolated_nodes.items():
        msg += "\n{} '{}' nodes are isolated:\n     {}\n".format(
            len(node_names), node_type, ", ".join(node_names)
        )
    print(msg)



# -*- coding: utf-8 -*-
import sys

from otoole.input import Context
from otoole.read_strategies import ReadCsv, ReadDatafile, ReadExcel, ReadMemory
from otoole.results.results import ReadCbc, ReadCplex, ReadGurobi
from otoole.write_strategies import WriteCsv, WriteDatafile, WriteExcel

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError, version  # pragma: no cover
else:
    from importlib_metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


__all__ = [
    "Context",
    "ReadCbc",
    "ReadCsv",
    "ReadCplex",
    "ReadDatafile",
    "ReadExcel",
    "ReadGurobi",
    "ReadMemory",
    "WriteCsv",
    "WriteDatafile",
    "WriteExcel",
]
